------------------------------------------------------------------------------
--
-- Simple OPB peripheral: a ETHENET controller
--
-- YASHKET GUPTA JAVIER COCA, ADAM
-- yg2113@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity sram_ethernet is

  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"00000000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"FFFFFFFF"
           );

  port (

    --ports between peripheral and OPB
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);  -- not using this
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;       -- Transfer acknowledge

       --ports between peripheral and Ethernet
    ETH_CS       : out std_logic;
    ETH_IOCS     : in  std_logic;
    ETH_RDY_DTAK : in  std_logic;
    ETH_IREQ     : in  std_logic;
    
 --    ETH_data     : inout std_logic_vector(15 downto 0)

     --ports between peripheral and SRAM
     -- IOR IOW AEN_PSEN BHE address and data
    RAM_CE          : out std_logic;
    PB_OE_IOR          : out std_logic;
    PB_WE_IOW          : out std_logic;
    PB_UB_BHE          : out std_logic;
    PB_LB_AEN          : out std_logic;
    PB_D   : inout std_logic_vector(0 to 15);
    PB_A   : out std_logic_vector(0 to 17)
    );
end sram_ethernet;

architecture Behavioral of sram_ethernet is

  
--deleted old signals from bram

      --SRAM constants
  constant C_SRAM_DWIDTH : integer                  := 16;
  constant C_SRAM_AWIDTH : integer                  := 18;

     -- Ethernet constants
  constant C_ETH_DWIDTH : integer                   := 16;
  constant C_ETH_AWIDTH : integer                   := 10;

     -- SRAM Process signals

  --signals from OPB
  -- universal signals same for both SRAM and Ethernet
  signal rnw, sel32 : std_logic;
  signal oe_ior, we_iow, ub_bhe, lb_aen : std_logic;
  signal data_in, data_out, tmp_data_in, tmp_data_out : std_logic_vector(0 to 15);
  signal address, tmp_address  : std_logic_vector(0 to 17);


--sram control signals
  signal RAMce , RAMcs : std_logic;  -- CS is test condition
                                                     -- CE is actual signal to
                                                     -- RAM
   --internal, used for fsm
  signal tristate : std_logic;

  --ethernet  control signals
    signal ETHcs, ETHce : std_logic;

  -- ethernet input signals
  signal ETHiocs, ETHireq, ETHrdy_dtak : std_logic;


  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS           : integer := 4;
  constant Idle                 : std_logic_vector(0 to STATE_BITS-1) := "0000";
  constant Select_Ram           : std_logic_vector(0 to STATE_BITS-1) := "0001";
  constant Select_Ethernet      : std_logic_vector(0 to STATE_BITS-1) := "0010";
  constant Read_RAM_ublb        : std_logic_vector(0 to STATE_BITS-1) := "0011";
  constant write_RAM_we         : std_logic_vector(0 to STATE_BITS-1) := "0100";
  constant Read_ETH             : std_logic_vector(0 to STATE_BITS-1) := "0101";
  constant Write_ETH            : std_logic_vector(0 to STATE_BITS-1) := "0110";
  constant Write_ETH_wait       : std_logic_vector(0 to STATE_BITS-1) := "0111";
  constant Xfer                 : std_logic_vector(0 to STATE_BITS-1) := "1000";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

  component OBUF_F_24
    port (
      O : out std_ulogic;
      I : in std_ulogic);
  end component;

  component IOBUF_F_24
    port (
      O : out std_ulogic;
      IO : inout std_ulogic;
      I : in std_ulogic;
      T : in std_ulogic);
  end component;

begin

-- IO Pads for the SRAM

   ADDR_PADS : for m in 0 to 17 generate   -- C_SRAM_AWIDTH-1 generate
    begin
      addr_pad : component OBUF_F_24 port map (
        I => address(m),
        O => PB_A(m));
    end generate ADDR_PADS;


    RAM_CE_PAD :OBUF_F_24 port map(
      O => RAM_CE,
      I => RAMce);

    ETH_CE_PAD :OBUF_F_24 port map(
      O => ETH_CS,
      I => ETHce);

    PB_OE_PAD :OBUF_F_24 port map(
      O => PB_OE_IOR,
      I => oe_ior);

    PB_WE_PAD :OBUF_F_24 port map(
      O => PB_WE_IOW,
      I => we_iow);

    PB_UB_PAD :OBUF_F_24 port map(
      O => PB_UB_BHE,
      I => ub_bhe);

    PB_LB_PAD :OBUF_F_24 port map(
      O => PB_LB_AEN,
      I => lb_aen);



    DATA_PADS : for m in 0 to 15 generate -- C_SRAM_DWIDTH-1 generate
    begin
      addr_pad : component IOBUF_F_24 port map (
        I => data_in(m),
        T => tristate,
        IO => PB_D(m),
        O => data_out(m));
    end generate DATA_PADS;

 -- IO Pads for the ethernet


-- internal chip select signals to trigger the FSM

--   RAMcs <=
--     '1' when OPB_select = '1' and
--         OPB_ABus(0 to C_OPB_AWIDTH-3-C_SRAM_AWIDTH) =
--         C_BASEADDR(0 to C_OPB_AWIDTH-3-C_SRAM_AWIDTH) else
--     '0';

--   ETHcs <=
--     '1' when OPB_select = '1' and
--         OPB_ABus(0 to C_OPB_AWIDTH - 1 - C_ETH_AWIDTH) =
--         C_BASEADDR(0 to C_OPB_AWIDTH- 1- C_ETH_AWIDTH) else
--     '0';

    ETHcs <=
      '1' when OPB_select = '1' and
          OPB_ABus(19) = '1' else
      '0';

    RAMcs <=
      '1' when OPB_select = '1' and
          OPB_ABus(19) = '0' else
      '0';

    sel32 <=
       '1' when ((OPB_BE = "0001") or (OPB_BE = "0010") or (OPB_BE = "0011")) else
       '0' when ((OPB_BE = "0100") or (OPB_BE = "1000") or (OPB_BE = "1100")) ;



-- Selecting the data

   with sel32 select

   tmp_data_in <=
   	OPB_DBus(0 to 15) when '1',
	OPB_DBus(16 to 31) when '0',
	"XXXXXXXXXXXXXXXX" when others;

  
-- conditional I/O process for "B O T H"

  ram_eth_opb_inputs: process (OPB_Clk, OPB_Rst, RAMcs, ETHcs)
    begin
      if  OPB_Rst = '1' then
       data_in <= (others => '0');
       address <= (others => '0');
        rnw <= '0';
      elsif OPB_Clk'event and OPB_Clk = '1' then
	data_in <= tmp_data_in;
        rnw <= OPB_RNW;
        if RAMcs = '1' then
          tmp_address <= OPB_ABus(C_OPB_AWIDTH - 2 -(C_SRAM_AWIDTH-1) to C_OPB_AWIDTH-2);
	  if sel32 = '1' then
          tmp_address <= OPB_ABus(C_OPB_AWIDTH - 2 -(C_SRAM_AWIDTH-1) to C_OPB_AWIDTH-3) & (OPB_ABus(30) or sel32);
	  elsif (ETHcs = '1') then
          tmp_address <= OPB_ABus(C_OPB_AWIDTH - 1 - (C_ETH_AWIDTH) to C_OPB_AWIDTH -1) & "00000000";
        end if;
        address <= tmp_address;
      end if;
      end if;
  end process ram_eth_opb_inputs;


  ram_eth_opb_outputs: process (OPB_Clk, OPB_Rst, RAMcs, oe_ior, sel32)
  begin
    if OPB_Rst = '1' then
       	Sln_DBus(0 to C_SRAM_DWIDTH-1) <= (others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
       	if oe_ior = '0' then
		if sel32 = '0' then
        		Sln_DBus(16 to 31) <= data_out;
		elsif sel32 = '1' then
        		Sln_DBus(0 to 15) <=  data_out;
        	end if;
        else
        	Sln_DBus(0 to C_SRAM_DWIDTH-1) <= (others => '0');
        end if;
    end if;
  end process ram_eth_opb_outputs;


-- unused outputs

  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';
  Sln_DBus(C_SRAM_DWIDTH to C_OPB_DWIDTH-1) <= (others => '0');  -- higher bits
                                                                 -- to zero

  --changed RAM_DWIDTH TO SRAM_DWIDTH


  --in the below we changed RAM_WIDTHS TO SRAM_WIDTHS
  -- we changed it Awidth to ETH_Awidth


  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= Idle;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(OPB_Rst, present_state, ETHcs, RAMcs, OPB_Select, RNW)
  begin

    -- Default values
    tristate <= '1';
    we_iow   <= '1';
    ETHce    <= '1';
    RAMce    <= '1';
    oe_ior   <= '1';
    lb_aen   <= '1';
    ub_bhe   <= '1';


    Sln_xferAck <= '0';

    if OPB_Rst = '1' then
      next_state <= Idle;
    else

      case present_state is

        when Idle =>
                if RAMcs = '1' then
                  next_state <= Select_Ram;
                elsif ETHcs = '1' then
                  next_state <= Select_Ethernet;
                else
                  next_state <= Idle;
                end if;

        when Select_Ram =>
                if OPB_Select = '1' then
		  if rnw = '1' then                           --READING
                    we_iow <= '1';         -- WE' is set to 1
                    RAMce <= '0';          -- CE' is set LOW
                    oe_ior <= '0';         -- OE' is set LOW
                    next_state <= Read_RAM_ublb;
                  else                                        --WRITING
                    RAMce <= '0'; -- WE' is set LOW
                     lb_aen <= NOT (OPB_BE(3) or OPB_BE(1));
		     ub_bhe <= NOT (OPB_BE(2) or OPB_BE(0));
                     next_state <= write_RAM_we;
                  end if;
                else
                  next_state <= Idle;
                end if;

        when Read_RAM_ublb =>
                if OPB_Select = '1' then
		  RAMce <= '0';
                  oe_ior <= '0';        --OE' is kept in LOW
                  lb_aen <= NOT (OPB_BE(3) or OPB_BE(1));
		  ub_bhe <= NOT (OPB_BE(2) or OPB_BE(0));
		  next_state <= Xfer;
                else
                  next_state <= Idle;
                end if;

        when write_RAM_we =>
                if OPB_Select = '1' then
                  we_iow <= '0';          -- CE' is set LOW for WRITING
                  lb_aen <= NOT (OPB_BE(3) or OPB_BE(1));
		  ub_bhe <= NOT (OPB_BE(2) or OPB_BE(0));
                  RAMce <= '0';
                  tristate <= '0';
                  next_state <= Xfer;
                else
                  next_state <= Idle;
                end if;

-- ethenet FSM

        when Select_Ethernet =>
                if OPB_Select = '1' then
		    ETHce <= '0';
		    lb_aen <= '1';       -- AEN is active high
                    ub_bhe <= '0';            -- Bit High Enable' is set LOW
                    if rnw = '1' then               --READING from ETHERNET
                       next_state <= Read_ETH;
                    else                            --WRITING to the ETHERNET
                      next_state <= Write_ETH;
                  end if;
                else
                  next_state <= Idle;
                end if;

        when Read_ETH=>
                if OPB_Select = '1' then
                  lb_aen <= '1';     -- AEN_PSEN' is set High
                  ub_bhe <= '0';          -- Bit High Enable' is set LOW
                  oe_ior <= '0';          -- IO Read' is set LOW for READING
                  next_state <= Xfer;
                else
                  next_state <= Idle;
                end if;

        when write_ETH=>
                if OPB_Select = '1' then
                  lb_aen <= '1';     -- AEN_PSEN' is set LOW
                  ub_bhe <= '0';          -- Bit High Enable' is set LOW
                  we_iow <= '0';          -- IO Write' is set LOW for READING
                  next_state <=write_ETH_wait;
                else
                  next_state <= Idle;
                end if;

        when write_ETH_wait=>              -- Waits for more then 5ns
                if OPB_Select = '1' then  -- everything is the same
                  lb_aen <= '0';     -- AEN_PSEN' is set LOW
                  ub_bhe <= '0';          -- Bit High Enable' is set LOW
                  we_iow <= '0';          -- IO Write' is set LOW for READING
                  next_state <=Xfer;
                else
                  next_state <= Idle;
                end if;

     -- State encoding is critical here: xfer must only be true here
        when Xfer =>            -- Actual Xfer
                Sln_xferAck <= '1';
                next_state <= Idle;

        when others =>
                next_state <= Idle;

      end case;
    end if;

  end process fsm_comb;


  
end Behavioral;

-- Local Variables:
-- compile-command: "ghdl -a opb_bram.vhd"
-- End:
